=begin
Copyright 2008-2009, Asteronimo (asteronimo@gmail.com)

Permission to use, copy, modify, and distribute this software for
any purpose and without fee is hereby granted, provided the above
copyright notice appears in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#-----------------------------------------------------------------------
  IF YOU FIND THIS SCRIPT AT ALL USEFUL I'D WELCOME YOUR DONATION AT:
  http://www.public-art-international.com/catalog/product_info.php/products_id/200
#-----------------------------------------------------------------------

Name        : SliceModeler.rb

Description : A tool to 'Slice' up a volume along 2 axes - useful for 
              rapid prototyping of real 3D models out of sheet material 
              like 3mm MDF or plexiglas.
              
              The 'thickness' parameter is used to determine the thickness
              of the slices and hence the width of the slot where 2 slices 
              interlock. 
              
Author      : Asteronimo (asteronimo@gmail.com)

Menu Item   : Plugins -> SliceModeler

Date        : 01/2009

Type        : Tool

Version     :
1.0 23/12/08  First working version based on TIG's Slicer script but without 
              transforming the slots to flattened slices.
1.1 27/12/08  Solved bug in slot position calculation
1.2 27/12/08  Using additional slices to calculate slot holes.
			  Improved parameter dialog and transforming slots to flattened slices.
1.3 03/01/09  Discovered that text labels add size to to bounding box dimensions.
              Corrected text labeling.
  			  Corrected contour cutting.
              Slices now keep their attributed thickness.
              Improved code for slicing in Z-direction and spreading the
              flattened slices. 
1.4 03/01/09  Moved slicing disc used to calculate intersections to outer loop.
              Solved issue with slot holes in contour with many edges.
              Restriction: A slot hole cannot have more than 7 edges.
1.4.1 06/01   Removed restriction.
              Improved speed of slot holes calculation.               
#-----------------------------------------------------------------------
=end


require 'sketchup.rb'

class SliceModeler

    ##################################################################################################
    #
    # --- Function that starts it all off
    #
    ##################################################################################################
    def SliceModeler::calculate

        @logging = false  # logging of intersections on/off
        @tl_size = 0.0.mm # text label size
                
        if (Sketchup.version.split(".")[0].to_i<5)
             UI.beep
             UI.messagebox("Sorry.  This is Version 5+ program only !")
             return nil
        end #if

        model = Sketchup.active_model
        model.active_layer = nil
        layer = nil
        entities = model.entities
        definitions = model.definitions
        model.start_operation "SliceModeler"
        ss = model.selection
        view = model.active_view
        sfix = Time::now.to_i.to_s[3..-1]

        if ss.empty?
           UI.beep
           UI.messagebox("NO Selection !")
           return nil
        end #if

        if ss[1]
           UI.beep
           UI.messagebox("Selection MUST be ONE Group or Component !")
           return nil
        end #if

        if ss[0].typename != "Group" and ss[0].typename != "ComponentInstance"
           UI.beep
           UI.messagebox("Selection is NOT a Group or Component !")
        end #if

        selected = ss[0]

        pslice      = []
        arr         = []
        xyz_setting = []

        @nn = 1
        while @nn < 3
            Sketchup::set_status_text(("SLICEMODELER PARAMETERS..." ), SB_PROMPT)
            Sketchup::set_status_text(" ", SB_VCB_LABEL)
            Sketchup::set_status_text(" ", SB_VCB_VALUE)
            return nil if !SliceModeler.dialog

            xyz_setting[@nn] = @xyz

            pslice[@nn]      = entities.add_group
            psliceentities   = pslice[@nn].entities
            pslice[@nn].name = "Slice-" + @xyz + "-" + sfix

            @sbb = selected.bounds
            xmin = @sbb.min.x
            xmax = @sbb.max.x
            ymin = @sbb.min.y
            ymax = @sbb.max.y
            zmin = @sbb.min.z
            zmax = @sbb.max.z

            if @xyz == "X"
	            pnt1 = [xmin,ymin,zmin] 
	            pnt2 = [xmin,ymax,zmax]
	            dist = ([xmin,0,0].distance [xmax,0,0])
            elsif @xyz == "Y"
	            pnt1 = [xmin,ymin,zmin] 
	            pnt2 = [xmax,ymin,zmax]
	            dist = ([0,ymin,0].distance [0,ymax,0])
            elsif @xyz == "Z"
	            pnt1 = [xmin,ymin,zmin] 
	            pnt2 = [xmax,ymax,zmin]
	            dist = ([0,0,zmin].distance [0,0,zmax])
            end #if
            
            spacing = @spacing

            if @nn == 1
               Sketchup::set_status_text(("ADD A NUMBER TO EACH SLICE ?" ), SB_PROMPT)
               Sketchup::set_status_text(" ", SB_VCB_LABEL)
               Sketchup::set_status_text(" ", SB_VCB_VALUE)

               @num = 1
               @num = 0 if UI.messagebox("Add a Number to each Slice ?   ",MB_YESNO,"Number ?")==7 ##=NO
            end #if

            q = (dist/(@spacing + @thickness)).floor
            myqdist = q * (@spacing + @thickness)

            if dist - myqdist < 4 * @thickness
               q -= 1
               myqdist = q * (@spacing + @thickness)
            end #if
            offset = (dist - myqdist + @spacing + @thickness)/2

            UI.messagebox("distance:" + dist.to_s + "\nspacing: " + @spacing.to_s + "\nthickness: " + @thickness.to_s + "\nnumber: " + q.to_s + "\noffset: " + offset.to_s)

            q += 1

            if q > 15 
               Sketchup::set_status_text(("LARGE NUMBER OF SLICES ? " ), SB_PROMPT)
               Sketchup::set_status_text(" ", SB_VCB_LABEL)
               Sketchup::set_status_text(" ", SB_VCB_VALUE)
               (return nil if UI.messagebox("This will make #{q.to_s} slices !   \nDo you want to continue ?   ",MB_YESNO,"Slice Count !")==7)##=NO
            end #if

            @ctr = Geom.linear_combination(0.5,pnt1,0.5,pnt2)
            @rad = (pnt1.distance pnt2) 
            ###
            disc = entities.add_group
            discentities = disc.entities
            if @xyz == "X"
               disccircle = SliceModeler.circle(@ctr,[-1,0,0],@rad,24) 
            elsif @xyz == "Y" 
               disccircle = SliceModeler.circle(@ctr,[0,-1,0],@rad,24)
            elsif @xyz == "Z"
               disccircle = SliceModeler.circle(@ctr,[0,0,-1],@rad,24)
            end #if
            #
            discentities.add_face disccircle

            @recurse_flag = false
            @hidden_flag = true

            t = @thickness * 1.5 
            m = spacing * 1      
            s = offset           
            slices = []

            #-------------------------------------------------------------------------------------------
            c = 1
            while c < q
                slice = psliceentities.add_group 
                sliceentities = slice.entities
                slice.name = "Slice-" + c.to_s

                Sketchup::set_status_text(("MAKING SLICE EDGES #{c} of #{q}"), SB_PROMPT)
                Sketchup::set_status_text(" ", SB_VCB_LABEL)
                Sketchup::set_status_text(" ", SB_VCB_VALUE)

                if @xyz == "X"
                   disc.move! Geom::Transformation.new [s,0,0] 
                elsif @xyz == "Y"
                   disc.move! Geom::Transformation.new [0,s,0] 
                elsif @xyz == "Z"
                   disc.move! Geom::Transformation.new [0,0,s] 
				end #if
				
                discentities.intersect_with @recurse_flag, disc.transformation, slice, slice.transformation, @hidden_flag, selected
                slice.layer = layer if slice
                ###
                slices = slices.push(slice) 
                ###
                s += spacing + @thickness
                
                c += 1
            end #while

            disc.erase!

            j = slices.length
            k = 1
            area = 0

            ###
            for slice in slices
                Sketchup::set_status_text(("PROCESSING SLICES #{k} of #{j}" ), SB_PROMPT)
                Sketchup::set_status_text(" ", SB_VCB_LABEL)
                Sketchup::set_status_text(" ", SB_VCB_VALUE)

                sliceentities = slice.entities

                for e in sliceentities
                   e.find_faces if e.typename == "Edge"
                end #for e

                faces = []
                for f in sliceentities
                   faces = faces.push(f) if f.typename == "Face"
                end #for f

                faces = faces.uniq

                for face in faces
                   if  @xyz == "X"   
                      face.reverse! if face.normal == [-1,0,0] 
                   elsif  @xyz == "Y"
                      face.reverse! if face.normal == [0,-1,0] 
                   elsif  @xyz == "Z"
                      face.reverse! if face.normal == [0,0,-1] 
                   end #if   
                end #for 

                (faces.length-1).times do |this|
                    for face in faces
                       if face.valid?
                          for edgeuse in face.outer_loop.edgeuses
                             if not edgeuse.partners[0] ### outermost face
                                faces = faces - [face]
                                loops = face.loops
                                for loop in loops
                                   for fac in faces
                                      if fac.valid? and (fac.outer_loop.edges - loop.edges) == []
                                         faces = faces - [fac]
                                         fac.erase! 
                                      end #if fac
                                   end #for fac
                                end #for loop
                             end #if outermost
                          end #for edgeuse
                       end #if valid
                    end #for face
                end #times

                faces = []
                for f in sliceentities
                    faces = faces.push(f) if f.typename == "Face"
                end #for f

                for f in faces
                   if @material == "<Default>"
                      f.material = nil
                   else
                      f.material = @material
                   end #if

                   f.layer = layer
                   for e in f.edges
                      e.layer = layer
                   end #for e
                   area += f.area
                end #for f

                for f in faces
                   f.reverse! if @thickness == 0 and @xyz == "Y"
                end #for f

                Sketchup::set_status_text((" " ), SB_PROMPT)
                Sketchup::set_status_text(" ", SB_VCB_LABEL)
                Sketchup::set_status_text(" ", SB_VCB_VALUE)

                k += 1
            end #for

            if area == 0
               UI.beep
               UI.messagebox("There is NO Volume to Slice !")
               pslice[@nn].erase! if pslice[@nn].valid?
               return nil
            end #if

            layerSlice = model.layers.add(@layerSlice)
            layerSlice.visible = true
            pslice[@nn].layer = layerSlice

            model.selection.clear

            xray = 0
            if not model.rendering_options["ModelTransparency"] ### NOT xray mode
               selected.hidden = true
               if UI.messagebox("Do you want to leave the Original Selection 'Hidden' for now ?  ",MB_YESNO,"Volume Hidden ?") == 7
                  xray = 1
                  selected.hidden = false
               end #if
            end #if

            xrayq = 0
            if not model.rendering_options["ModelTransparency"]
               xrayq = 1
               model.rendering_options["ModelTransparency"]=true if xray == 1
            end #if

            if xray == 1 and xrayq == 1
               model.rendering_options["ModelTransparency"]=false if UI.messagebox("Do you want to leave 'Xray Mode' on for now ?  ",MB_YESNO,"Volume Xray ?") == 7 ###7=NO
            end #if

            @nn += 1
        end #while

        holefaces = []

        Sketchup::set_status_text(("FINDING EDGES OF INTERSECTIONS... " ), SB_PROMPT)
        Sketchup::set_status_text(" ", SB_VCB_LABEL)
        Sketchup::set_status_text(" ", SB_VCB_VALUE)
        
        rr = 1
        while rr < 3

            if rr == 1
               aaa = 1
               bbb = 2
            else
               aaa = 2
               bbb = 1
            end #if

            for l in pslice[aaa].entities
               #
               kp  = []
               if xyz_setting[bbb] == "Z"
	              kp = kp.push([@sbb.min.x,@sbb.min.y,@sbb.min.z])    
                  kp = kp.push([@sbb.min.x,@sbb.max.y,@sbb.min.z])    
                  kp = kp.push([@sbb.max.x,@sbb.max.y,@sbb.min.z])    
                  kp = kp.push([@sbb.max.x,@sbb.min.y,@sbb.min.z])    
               elsif xyz_setting[bbb] == "Y"
	              kp = kp.push([@sbb.min.x,@sbb.min.y,@sbb.min.z])    
                  kp = kp.push([@sbb.max.x,@sbb.min.y,@sbb.min.z])    
                  kp = kp.push([@sbb.max.x,@sbb.min.y,@sbb.max.z])    
                  kp = kp.push([@sbb.min.x,@sbb.min.y,@sbb.max.z])    
               elsif xyz_setting[bbb] == "X"
	              kp = kp.push([@sbb.min.x,@sbb.min.y,@sbb.min.z])    
                  kp = kp.push([@sbb.min.x,@sbb.max.y,@sbb.min.z])    
                  kp = kp.push([@sbb.min.x,@sbb.max.y,@sbb.max.z])    
                  kp = kp.push([@sbb.min.x,@sbb.min.y,@sbb.max.z])    
               end #if
              
               dsc = entities.add_group 
               dscentities = dsc.entities
               dscentities.add_face kp
                 
               for k in pslice[bbb].entities
	              ###
                  kb   = k.bounds
	              xoff = kb.min.x - @sbb.min.x
	              yoff = kb.min.y - @sbb.min.y
	              zoff = kb.min.z - @sbb.min.z
	              #
		          p_l = []
		          p_c = []
		          p_r = []
	              
                  slice_repository = []
                  slicepos = -@thickness/2
                  #
                  tmpcnt = 0
                  #
                  1.upto(3) { |num|
                      tmp_slices = []
                      #
	                  tmpgroup = entities.add_group 
	                  
	                  if xyz_setting[bbb] == "Z"
	                     dsc.move! Geom::Transformation.new [0,0,zoff+slicepos]
	                  elsif xyz_setting[bbb] == "Y"
	                     dsc.move! Geom::Transformation.new [0,yoff+slicepos,0]
	                  elsif xyz_setting[bbb] == "X"
	                     dsc.move! Geom::Transformation.new [xoff+slicepos,0,0]
	                  end #if
	                  #
	                  dscentities.intersect_with true, dsc.transformation, tmpgroup, tmpgroup.transformation, true, l
	                  #
                      tmpgroup.entities.each { |e|
                         tmp_slices = tmp_slices.push([e.start.position,e.end.position])
                         tmpcnt += 1
                      } #each
	                  
	                  slice_repository = slice_repository.push(tmp_slices)
	                  #
	                  tmpgroup.erase!
	                  #
	                  slicepos += @thickness/2
                  }
                  
                  if tmpcnt > 0 and (tmpcnt % 3) == 0 
	                  if xyz_setting[bbb] == "Z"
	                     p_l = SliceModeler.xsort(slice_repository.shift)
	                     p_c = SliceModeler.xsort(slice_repository.shift)
	                     p_r = SliceModeler.xsort(slice_repository.shift)
	                  else
	                     p_l = SliceModeler.zsort(slice_repository.shift)
	                     p_c = SliceModeler.zsort(slice_repository.shift)
	                     p_r = SliceModeler.zsort(slice_repository.shift)
	                  end #if
                      #
                      c = tmpcnt/3
                      #
                      pts = []
                      #
                      if xyz_setting[bbb] == "X"
                          if xyz_setting[aaa] == "Y"
                              0.upto(c-1) { |ccnt|
                   	             zm = (p_c[ccnt][1].z + p_c[ccnt][0].z)/2
                                 pts[0] = Geom::Point3d.new([p_l[ccnt][0].x,p_l[ccnt][0].y,p_l[ccnt][0].z])
                                 pts[1] = Geom::Point3d.new([p_l[ccnt][0].x,p_l[ccnt][0].y,zm])
                                 pts[2] = Geom::Point3d.new([p_r[ccnt][0].x,p_r[ccnt][0].y,zm])
                                 pts[3] = Geom::Point3d.new([p_r[ccnt][0].x,p_r[ccnt][0].y,p_r[ccnt][0].z])
                                 #
                                 SliceModeler.log_intersection(ccnt,pts) if @logging
		                         holefaces = holefaces.push([rr,l,pts.dup])
                              }
                          elsif xyz_setting[aaa] == "Z"
                              0.upto(c-1) { |ccnt|
                   	             zm = (p_c[ccnt][1].y + p_c[ccnt][0].y)/2
                                 pts[0] = Geom::Point3d.new([p_l[ccnt][1].x,p_l[ccnt][1].y,p_l[ccnt][1].z])
                                 pts[1] = Geom::Point3d.new([p_l[ccnt][1].x,zm,            p_l[ccnt][1].z])
                                 pts[2] = Geom::Point3d.new([p_r[ccnt][1].x,zm,            p_r[ccnt][1].z])
                                 pts[3] = Geom::Point3d.new([p_r[ccnt][1].x,p_r[ccnt][1].y,p_r[ccnt][1].z])
                                 #
                                 SliceModeler.log_intersection(ccnt,pts) if @logging
		                         holefaces = holefaces.push([rr,l,pts.dup])
                              }
                          end #if
                      elsif xyz_setting[bbb] == "Y"
                          if xyz_setting[aaa] == "X"
                              0.upto(c-1) { |ccnt|
                   	             zm = (p_c[ccnt][1].z + p_c[ccnt][0].z)/2
                                 pts[0] = Geom::Point3d.new([p_l[ccnt][1].x,p_l[ccnt][1].y,p_l[ccnt][1].z])
                                 pts[1] = Geom::Point3d.new([p_l[ccnt][1].x,p_l[ccnt][1].y,zm])
                                 pts[2] = Geom::Point3d.new([p_r[ccnt][1].x,p_r[ccnt][1].y,zm])
                                 pts[3] = Geom::Point3d.new([p_r[ccnt][1].x,p_r[ccnt][1].y,p_r[ccnt][1].z])
                                 #
                                 SliceModeler.log_intersection(ccnt,pts) if @logging
		                         holefaces = holefaces.push([rr,l,pts.dup])
	                          }
	                      elsif xyz_setting[aaa] == "Z"
	                          0.upto(c-1) { |ccnt|
                   	             zm = (p_c[ccnt][1].x + p_c[ccnt][0].x)/2
	                             pts[0] = Geom::Point3d.new([p_l[ccnt][1].x,p_l[ccnt][1].y,p_l[ccnt][1].z])
	                             pts[1] = Geom::Point3d.new([zm,            p_l[ccnt][1].y,p_l[ccnt][1].z])
	                             pts[2] = Geom::Point3d.new([zm,            p_r[ccnt][1].y,p_r[ccnt][1].z])
	                             pts[3] = Geom::Point3d.new([p_r[ccnt][1].x,p_r[ccnt][1].y,p_r[ccnt][1].z])
                                 #
                                 SliceModeler.log_intersection(ccnt,pts) if @logging
		                         holefaces = holefaces.push([rr,l,pts.dup])
                              }
                          end #if
                      elsif xyz_setting[bbb] == "Z"
                          if xyz_setting[aaa] == "X"
                              0.upto(c-1) { |ccnt|
                   	             zm = (p_c[ccnt][1].y + p_c[ccnt][0].y)/2
                                 pts[0] = Geom::Point3d.new([p_l[ccnt][0].x,p_l[ccnt][0].y,p_l[ccnt][0].z])
                                 pts[1] = Geom::Point3d.new([p_l[ccnt][0].x,zm,            p_l[ccnt][0].z])
                                 pts[2] = Geom::Point3d.new([p_r[ccnt][0].x,zm,            p_r[ccnt][0].z])
                                 pts[3] = Geom::Point3d.new([p_r[ccnt][0].x,p_r[ccnt][0].y,p_r[ccnt][0].z])
                                 #
                                 SliceModeler.log_intersection(ccnt,pts) if @logging
		                         holefaces = holefaces.push([rr,l,pts.dup])
                              }
                          elsif xyz_setting[aaa] == "Y"
                              0.upto(c-1) { |ccnt|
                   	             zm = (p_c[ccnt][1].x + p_c[ccnt][0].x)/2
	                             pts[0] = Geom::Point3d.new([p_l[ccnt][0].x,p_l[ccnt][0].y,p_l[ccnt][0].z])
	                             pts[1] = Geom::Point3d.new([zm,            p_l[ccnt][0].y,p_l[ccnt][0].z])
	                             pts[2] = Geom::Point3d.new([zm,            p_r[ccnt][0].y,p_r[ccnt][0].z])
	                             pts[3] = Geom::Point3d.new([p_r[ccnt][0].x,p_r[ccnt][0].y,p_r[ccnt][0].z])
                                 #
                                 SliceModeler.log_intersection(ccnt,pts) if @logging
		                         holefaces = holefaces.push([rr,l,pts.dup])
                              }
                          end #if
                      end # if
                  end #if
               end #for
                  
               dsc.erase!
               
            end #for
            
            rr += 1
        end #while rr
               
        #-------------------------------------------------------------------------------------------

        Sketchup::set_status_text(("MAKING SLOT HOLES... " ), SB_PROMPT)
        Sketchup::set_status_text(" ", SB_VCB_LABEL)
        Sketchup::set_status_text(" ", SB_VCB_VALUE)

        1.upto(2) { |k|
           c = 1
	       pslice[k].entities.each { |l|
	          holefaces.each { |n,e,p|
	             if n == k and l == e 
			        edges = l.entities.add_edges p
			        if edges[0].faces.length > 0
				        f = edges[0].faces[0] 
				        del_edges = []
				       	f.edges.each { |fe|
		                   del_edges = del_edges.push(fe) if fe.faces.length == 1
				        } #each
			            f.erase!
	                    del_edges.each { |d|
	                       begin
	                          l.entities.erase_entities d 
	                       rescue => e
	                          #puts e.message
	                       end   
	                    }
                    end #if
		         end #if
	          }
              l.entities.each { |e|
                 if e.typename == "Edge" 
                    e.erase! if e.faces == 0
                 end #if 
              } #each
              ###
              if @num == 1
	              bb = l.bounds
	              zmin = bb.min.z
	              ymin = bb.min.y
	              xmin = bb.min.x
	              ###
	              if xyz_setting[k] == "X"
	                 tt = l.entities.add_text(xyz_setting[k].to_s + c.to_s, [xmin+t,ymin+m,zmin+m]) 
	              elsif xyz_setting[k] == "Y"
	                 tt = l.entities.add_text(xyz_setting[k].to_s + c.to_s, [xmin+m,ymin+t,zmin+m]) 
	              elsif xyz_setting[k] == "Z"
	                 tt = l.entities.add_text(xyz_setting[k].to_s + c.to_s, [xmin+m,ymin+m,zmin+t]) 
	              end #if
	              ###
	              tt.layer = layer if tt
	              #
	              @tl_size = (tt.bounds.max.y - tt.bounds.min.y)/2

              end #if
              c += 1
	       } #each
        } #upto
        
        if @thickness != 0
	        if xyz_setting[1] == "Z"
	           if xyz_setting[2] == "X"
		          pslice[1].move! Geom::Transformation.new [@thickness/2,0,0]
		          pslice[2].move! Geom::Transformation.new [0,0,@thickness/2]
	           elsif xyz_setting[2] == "Y"
		          pslice[1].move! Geom::Transformation.new [0,@thickness/2,0]
		          pslice[2].move! Geom::Transformation.new [0,0,@thickness/2]
	           end #if
	        elsif xyz_setting[1] == "X"
	           if xyz_setting[2] == "Y"
		          pslice[1].move! Geom::Transformation.new [0,@thickness/2,0]
		          pslice[2].move! Geom::Transformation.new [@thickness/2,0,0]
	           elsif xyz_setting[2] == "Z"
		          pslice[1].move! Geom::Transformation.new [0,0,@thickness/2]
		          pslice[2].move! Geom::Transformation.new [@thickness/2,0,0]
		       end #if
	        elsif xyz_setting[1] == "Y"
	           if xyz_setting[2] == "X"
		          pslice[1].move! Geom::Transformation.new [@thickness/2,0,0]
		          pslice[2].move! Geom::Transformation.new [0,@thickness/2,0]
	           elsif xyz_setting[2] == "Z"
		          pslice[1].move! Geom::Transformation.new [0,0,@thickness/2]
		          pslice[2].move! Geom::Transformation.new [0,@thickness/2,0]
		       end #if   
	        end #if
		    #
	        1.upto(2) { |k|
		       pslice[k].entities.each { |l|
	              faces = []
	              for f in l.entities
	                  faces = faces.push(f) if f.typename == "Face"
	              end #for f
	              for f in faces
	                 (f.pushpull @thickness)
	                 f.reverse! if @xyz == "Y"
	              end #for f
		       }
	        } 
        end #if
        
        if UI.messagebox("Do you want to make 'Flattened' copies of the Slices ?  ",MB_YESNO,"Flatten ?") == 6 ### 6=YES 7=NO

           if SliceModeler.gap 

             Sketchup::set_status_text(("FLATTENING SLICES..." ), SB_PROMPT)
             Sketchup::set_status_text(" ", SB_VCB_LABEL)
             Sketchup::set_status_text(" ", SB_VCB_VALUE)

             gap = @gap

             ypos = 0
             
             @nn = 1
             while @nn < 3
                fslice = pslice[@nn].copy
                fslice.name = "Flat-" + xyz_setting[@nn].to_s + "-" + sfix
                fents = fslice.entities
                fs = fents.length
                fb = pslice[@nn].bounds
                z = fb.min.z
                y = fb.min.y
                x = fb.min.x
                if xyz_setting[@nn] == "Z"
                   y  = fb.max.y - fb.min.y
                else
                   ym = fb.max.z - fb.min.z
                end #if z

                if xyz_setting[@nn] == "X"
                   xm = fb.max.z - fb.min.z
                   yg = fb.max.y - fb.min.y
                   xs = x
                end #if x
                xx = 0
                yy = 0
                sq = Math::sqrt(fs).to_i 
                sqq = sq
                c = 0

                for e in fents
                   e.layer = layer
                   p = e.bounds.min
                   px = p.x
                   py = p.y
                   pz = p.z
                   pp = e.bounds.max
                   pxx = pp.x
                   pyy = pp.y
                   pzz = pp.z
                   #
                   if xyz_setting[@nn] == "X"
                      pt = [xx,y - py,z - pz]     
                   elsif xyz_setting[@nn] == "Y"
                      pt = [xx,y - py,z - pz]     
                   elsif xyz_setting[@nn] == "Z"
                      pt = [xx,yy,-pz - @tl_size] 
                   end #if
                   #
                   tr = Geom::Transformation.new pt
                   e.transform! tr

                   if xyz_setting[@nn] == "X"
                      xx = xx + xm + gap 
                   else
                      xx = xx + (pxx - px) + gap 
 				   end #if
 				   	
                   if c == sqq and xyz_setting[@nn] == "Z"
                      yy = yy + (y + gap)
                      xx = 0
                      sqq = sqq + sq + 1
                   elsif c == sqq and xyz_setting[@nn] != "Z"
                      if xyz_setting[@nn] == "Y"
                         y = y + (ym + gap) 
                      elsif xyz_setting[@nn] == "X"
                         y = y + (yg + gap) 
                      end #if
                      if xyz_setting[@nn] == "Y"
                         xx = 0 
                      elsif xyz_setting[@nn] == "X"
                         xx = xs - px 
                      end #if
                      sqq = sqq + sq + 1
                   end #if

                   c += 1
                end #for e

                tmp = fents.add_line fents[0].bounds.min,fents[fents.length - 1].bounds.max
                tmp.erase!

                if xyz_setting[@nn] != "Z"
                   r = Math::PI / 2
                   for e in fents
                      eb = e.bounds.min
                      p = [eb.x,eb.y,fb.min.z]
                      if xyz_setting[@nn] == "X"
                         v = Geom::Vector3d.new 0,-1,0 
                      elsif xyz_setting[@nn] == "Y"
                         v = Geom::Vector3d.new -1,0,0 
                      elsif xyz_setting[@nn] == "Z" ###
                         v = Geom::Vector3d.new 0,0,-1 
                      end #if
                      rotate = Geom::Transformation.rotation p,v,r
                      e.transform! rotate
                   end #for e
               end #if !=Z
                                  
                tmp = fents.add_line fents[0].bounds.min,fents[fents.length - 1].bounds.max
                tmp.erase!

                if @nn == 1
                   ypos = fslice.bounds.max.y
                else  
                   fslice.move! Geom::Point3d.new [0,@gap + ypos,0]
                end #if
                
                layerSlice = ("FLAT-" + @layerSlice) 
                layerSlice = model.layers.add(layerSlice)
                layerSlice.visible = true
                fslice.layer = layerSlice

                @nn += 1
             end #while

             Sketchup::set_status_text(("FLATTENING SLICES DONE." ), SB_PROMPT)
             Sketchup::set_status_text(" ", SB_VCB_LABEL)
             Sketchup::set_status_text(" ", SB_VCB_VALUE)

          end #if Gap
        end #if Flatten

        view.invalidate

        Sketchup::set_status_text(("DONE!" ), SB_PROMPT)
        Sketchup::set_status_text(" ", SB_VCB_LABEL)
        Sketchup::set_status_text(" ", SB_VCB_VALUE)

        model.commit_operation

    end #def

    ##################################################################################################
    #
    # --- SliceModeler::log_intersection(ccnt,pts)
    #
    ##################################################################################################
    def SliceModeler::log_intersection(ccnt,pts)
        puts "---- Intersection " + ccnt.to_s + "--------------------------"
        puts "0: " + pts[0].to_s + ", 1: " + pts[1].to_s
        puts "2: " + pts[2].to_s + ", 3: " + pts[3].to_s
        puts "------------------------------"
    end #def
    
    ##################################################################################################
    #
    # --- SliceModeler::xsort(ar)
    #
    ##################################################################################################
    def SliceModeler::xsort(ar)
        sortar = []
        #
        ar.each { |n|
           if n[0].x > n[1].x
              tmp = n[0]
              n[0] = n[1]
              n[1] = tmp
           end #if
        }
        #
        while ar.size > 0
           ns = 0
           1.upto(ar.size-1) { |n|
              if ar[n][1].x > ar[ns][1].x
                ns = n
              end #if
           }
           sortar = sortar.push(ar[ns])
           ar.delete_at(ns)
        end #while
        #
        return sortar
    end #def
    
    ##################################################################################################
    #
    # --- SliceModeler::zsort(ar)
    #
    ##################################################################################################
    def SliceModeler::zsort(ar)
        sortar = []
        #
        ar.each { |n|
           if n[0].z > n[1].z
              tmp = n[0]
              n[0] = n[1]
              n[1] = tmp
           end #if
        }
        #
        while ar.size > 0
           ns = 0
           1.upto(ar.size-1) { |n|
              if ar[n][1].z > ar[ns][1].z
                ns = n
              end #if
           }
           sortar = sortar.push(ar[ns])
           ar.delete_at(ns)
        end #while
        #
        return sortar
    end #def
    
    ##################################################################################################
    #
    # --- Dialog to enter spacing, thickness, layer and colour
    #
    ##################################################################################################
    def SliceModeler::dialog

        ### get current units ----------------------
        currentunits1 = Sketchup.format_length 120000
        currentunits2 = Sketchup.format_length 1
        units = "???" ### default

        cu = currentunits1.to_s.split(".")[0].split(",")[0].split("'")[0].split("\"")[0].split("c")[0].split("m")[0]
        if cu == "120000"
           units = "inches"
        elsif cu == "10000"
           units = "feet"
        elsif cu == "3048000"
           units = "mm"
        elsif cu == "304800"
           units = "cm"
        elsif cu == "3048"
           units = "m"
        end #if   
        if currentunits2 == "1\""
           units = "inches or \" "
        end #if

        if units[-1,1] == "m" 
           xspacing = 50.mm
           xthickness = 3.mm
        else ### not metric so these are in feet
           xspacing = 2.inch
           xthickness = 0.125.inch
        end #if

        mlayers = Sketchup.active_model.layers
        layers = []
        mlayers.each{|e|layers.push e.name}
        dlayer = layers[0]
        layers = layers - [dlayer]
        layers.sort!

        @sfix = Time::now.to_i.to_s[3..-1]
        layerSlice = ("SLCE-" + @sfix)
        @makelayer = "<Make New Layer>"
        layers = [dlayer] + [layerSlice] + [@makelayer] + layers
        layers.uniq!
        layers = layers.join('|')

        mcolours = Sketchup.active_model.materials
        colours = []
        mcolours.each{|e|colours.push e.name}
        colours.sort!
        colours = ["<Default>"] + colours + (Sketchup::Color.names.sort!)
        colours.uniq!
        colours = colours.join('|')

        @spacing    = xspacing    if not @spacing
        @thickness  = xthickness  if not @thickness
        @layerSlice = dlayer      if not @layerSlice
        @material   = "<Default>" if not @material

        if @nn == 2
            title = "SliceModeler - SLICE 2 Parameters"
	        if @xyz == "X"
               @xyz = "Y"
	           xyz = "Y|Z"
	        elsif @xyz == "Y"
               @xyz = "X"
	           xyz = "X|Z"
	        elsif @xyz == "Z"
               @xyz = "X"
	           xyz = "X|Y"
	        else
               @xyz = "X"
	           xyz = "X|Y|Z"
	        end #if
            prompts = ["Slice Spacing (" + units + "): ","Slice Orientation: ","Layer: ","Colour: "]
            types   = ["",xyz,layers,colours]
            values  = [@spacing,@xyz,@layerSlice,@material]
            popups  = ["",xyz,layers,colours]

	        results = inputbox( prompts, values, popups, title )
	        return nil if not results

	        @spacing    = results[0]
	        @xyz        = results[1]
	        @layerSlice = results[2]
	        @material   = results[3]
        else
            title = "SliceModeler - SLICE 1 Parameters"
            @xyx = "X"
	        xyz = "X|Y|Z"

            prompts = ["Slice Spacing (" + units + "): ","Slice Thickness (" + units + "): ","Slice Orientation: ","Layer: ","Colour: "]
            types   = ["","",xyz,layers,colours]
            values  = [@spacing,@thickness,@xyz,@layerSlice,@material]
            popups  = ["","",xyz,layers,colours]

	        results = inputbox( prompts, values, popups, title )
	        return nil if not results

	        @spacing    = results[0]
	        @thickness  = results[1]
	        @xyz        = results[2]
	        @layerSlice = results[3]
	        @material   = results[4]
        end #if

        if @spacing <= 0
           UI.beep
           UI.messagebox("Spacing CAN'T be <= 0 !  \nMade default = #{xspacing} #{units}")
           @spacing = xspacing
        end#if

        if @thickness > @spacing
           UI.beep
           if xthickness < @spacing
              UI.messagebox("Thickness CANNOT > Spacing !  \nThickness = default = #{xthickness} #{units}")
              @thickness = xthickness
           else
              UI.messagebox("Thickness CANNOT > Spacing !  \nThickness = Spacing = #{@spacing} #{units}")
              @thickness = @spacing
           end #if
        end #if

        if @layerSlice == @makelayer
           results2 = inputbox(["New Slices Layer Name: "],["????"],"New Layer Name")
           if results2
              @layerSlice = results2[0]
           else
              @layerSlice = nil 
              return nil 
           end #if
        end #if

        true
    end #def dialog

    ##################################################################################################
    #
    # --- Dialog to enter gap for flattened slices
    #
    ##################################################################################################
    def SliceModeler::gap

        ### get current units ----------------------
        currentunits1 = Sketchup.format_length 120000
        currentunits2 = Sketchup.format_length 1
        units = "???" 

        cu = currentunits1.to_s.split(".")[0].split(",")[0].split("'")[0].split("\"")[0].split("c")[0].split("m")[0]
        if cu == "120000"
           units = "inches"
        elsif cu == "10000"
           units = "feet"
        elsif cu == "3048000"
           units = "mm"
        elsif cu == "304800"
           units = "cm"
        elsif cu == "3048"
           units = "m"
        end #if   
        if currentunits2 == "1\""
           units = "inches or \" "
        end #if
        ### ---------------
        if units[-1,1] == "m" 
           xgap = 10.0.mm
        else 
           xgap = 0.5.inch
        end #if

        prompts = ["#{units}: "]
        types = [""]
        title = "Minimum Gap"
        @gap = xgap if not @gap
        values = [@gap]
        popups = [""]
        results = inputbox( prompts, values, popups, title )
        return nil if not results

        @gap = results[0]

        true
    end #def gap

    ##################################################################################################
    #
    # --- Function for generating points on a circle
    #
    ##################################################################################################
    def SliceModeler::circle(center,normal,radius,numseg)

        axes = Geom::Vector3d.new(normal).axes
        center = Geom::Point3d.new(center)
        xaxis = axes[0]
        yaxis = axes[1]
        xaxis.length = radius
        yaxis.length = radius

        da = (Math::PI * 2) / numseg
        pts = []
        for i in 0...numseg do
            angle = i * da
            cosa = Math.cos(angle)
            sina = Math.sin(angle)
            vec = Geom::Vector3d.linear_combination(cosa,xaxis,sina,yaxis)
            pts.push(center + vec)
        end

        pts.push(pts[0].clone)
        pts
    end #def

end #class


if( not file_loaded?("SliceModeler.rb") )
   add_separator_to_menu("Plugins")###
   UI.menu("Plugins").add_item("SliceModeler 1.4") { SliceModeler.calculate }###
   ### $submenu5.add_item("SliceModeler") { SliceModeler.calculate }###
   ### remove ### in front of 6 lines below if to display in context-menu...
   ### UI.add_context_menu_handler do | menu |
   ###   if Sketchup.active_model.selection[0].typename == "Group" or Sketchup.active_model.selection[0].typename == "ComponentInstance"
   ###      menu.add_separator
   ###      menu.add_item("SliceModeler 1.4") { SliceModeler.calculate }
   ###   end #if ok
   ### end #do menu
end#if

file_loaded("SliceModeler.rb")
